"""
Supports loading animations, inspired by Auto-GPT's spinner class, found here: https://github.com/Significant-Gravitas/Auto-GPT/blob/master/autogpt/spinner.py
This version supports customizable cycling as well as default classes for quick access. You can also print messages above the spinner.
"""
from __future__ import annotations
from threading import Thread
from time import sleep

class Loader():
    
    def __init__(self, message: str, cycle: list[str], delay: int, hide_cursor: bool = True) -> None:

        """
        Creates a new Loader with a message which appears on the same line as the cycler.

        :param str message: The message being printed on the same line as the cycler.
        :param list[str] cycle: The list of strings to cycle between to display loading.
        :param int delay: The number of milliseconds between each item in the cycle.
        :param bool hide_cursor: Whether or not the cursor is hidden while the Loader runs.

        :raises AssertionError: if either the message or the items in the cycle are not strings.

        :return: None
        :rtype: NoneType
        """

        assert all([isinstance(part, str) for part in cycle]), "Cycle must consist of strings only."
        assert isinstance(message, str), "Message must be a string."

        self.max_len = max([len(item) for item in cycle])

        # stores customization values
        self.cycle = cycle
        self.message = message
        self.delay = delay / 1000
        self.hide_cursor = hide_cursor

        # temporary null value 
        self.runner = None

        # stores values for going back and clearing
        self.prev_length = len(self.cycle[0]) 
        
        # for cycling
        self.mod = len(cycle)
        self.index = 0

        # for states
        self.over = False
        self.paused = False
        
        if self.hide_cursor:
            print("\033[?25l", end="")

    @property
    def __state(self):
        """
        Adjusts the index and returns it.
        """
        self.index += 1
        self.index %= self.mod
        return self.index

    def __run(self) -> None:
        """
        Thread that runs the loader.
        """
        try:
            while not self.over:
                if self.paused:
                    continue
                item = self.cycle[self.__state]
                self.__display(item)
                sleep(self.delay)
        finally:
            print("\033[?25h", end="")  # shows cursor no matter what

    def __display(self, item):
        """
        Clears and displays the message.
        """
        self.__clear_animated()
        self.prev_length = len(item)
        print("\r" + self.message + item, end="")

    def __enter__(self) -> Loader:
        self.runner = Thread(target=self.__run)
        self.runner.start()
        return self

    def __exit__(self, _, __, ___):

        # ends and clears
        self.over = True
        self.__clear()

        # print newline
        print(end="\r") 

        # end thread
        self.runner.join()

    def print_above(self, message) -> None:
        """
        Prints the entered string above the Loader.

        :param str message: The message to be printed above the Loader.
        """

        # pause and print message
        self.paused = True
        self.__clear()
        print("\r" + message)
        self.paused = False

        # reset the display
        self.__display(self.cycle[self.__state])
        sleep(self.delay)

    def __clear_animated(self):
        """
        Clears the animated thing.
        """
        print("\b" * self.prev_length 
              + " " * self.prev_length 
              + "\b" * self.prev_length, end="")

    def __clear(self):
        """
        Clears the message.
        """
        print("\r" + " " * (len(self.message) + self.max_len + 1), end="")

class Spinner(Loader):
    """
    A Spinner preset based on the Loader class.
    Cycles between \, |, /, and - to simulate spinning.
    """
    def __init__(self, message: str, delay: int = 50) -> None:
        """
        A Spinner preset based on the Loader class.
        
        :param str message: The message on the same line as the spinner.
        :param int delay: Delay in milliseconds between each item, defaulted to 50.

        :raises AssertionError: if the message is not a string.

        :return: None
        :rtype: NoneType
        """
        super().__init__(message=message, cycle=["/", "-", "\\", "|"], delay=delay, hide_cursor=True)

class Ellipsis(Loader):
    """
    A Ellipsis - "..." - preset based on the Loader class.
    Cycles up to a given number of periods then repeats.
    """
    def __init__(self, message: str, periods: int = 3, delay: int = 200) -> None:
        """
        A Ellipsis - "..." - preset based on the Loader class.
        
        :param str message: The message that will be printed next to the cycler.
        :param int periods: The number of periods at the longest string.
        :param int delay: Delay in milliseconds between each item, defaulted to 200.

        :raises AssertionError: if the message is not a string or if the periods are anything but an int greater than 1.

        :return: None
        :rtype: NoneType
        """
        assert isinstance(periods, int) and periods > 1, "You must have an upper limit of at least 2 periods!"
        super().__init__(message=message, cycle=["."*n for n in range(periods+1)], delay=delay, hide_cursor=True)

def test_loaders():
    """
    Tests the functionality and displays of loaders.
    """
    def __loop():
        """
        Example time-consuming task to test loaders.
        """
        for i in range(1, 100_000_000):
            if i % 40_000_000 == 0:
                yield f"{i:,} is divisble by 40,000,000!"
            elif i % 20_000_000 == 0:
                yield f"{i:,} is almost divisble by 40,000,000!"

    with Loader(message="Loading ", cycle=[str(n) for n in range(10)], delay=10) as loader:
        for output in __loop():
            loader.print_above(output)

    print("Loader done!")

    with Spinner(message="Calculating ") as spinner:
        for output in __loop():
            spinner.print_above(output)

    print("Spinner done!")

    with Ellipsis(message="Periods", periods=15) as ellipsis:
        for output in __loop():
            ellipsis.print_above(output)

    print("Ellipsis done!")

if __name__ == "__main__":
    test_loaders()


"""
A python module for dealing with terminal-based inputs and outputs.
"""

__version__ = "0.0.3"
__author__ = "Vivaan Singhvi"

from .menu import *
from .input import *
from .style import *
from .loading import *
from .printing import *


"""
Input tools, currently supported getting a bounded integer and a two-dimensional array.
"""

from typing import Any, Iterable

def bounded_int(lower: int, upper: int, prompt: str = None, 
                type_error: str = None, bounds_error: str = None) -> int:
    """
    Forces the user to return an integer between the two entered values.

    Arguments: A lower and an upper bound for the integers. An optional prompt.

    Raises: An AssertionError if the types of each the lower and upper bound are not integers, or is the upper bound si not greater than the lower

    Returns: The correct integer that the user inputs.
    """

    assert isinstance(lower, int), f"Lower bound must be an integer, not '{type(lower)}'."
    assert isinstance(upper, int), f"Upper bound must be an integer, not '{type(upper)}'."
    assert upper > lower, "Upper bound must be greater than lower bound."

    if not prompt:
        prompt = f"Please enter a number between {lower} and {upper}: "

    if not type_error:
        type_error = "You must enter an integer: "

    if not bounds_error:
        bounds_error = f"The number must be between {lower} and {upper}: "

    user_input = (lower - 1) / 1    # convert to float
    flag_type_error = False
    while not (lower <= user_input <= upper):
        try:
            if isinstance(user_input, float):       # get first input
                user_input = int(input(prompt))
            elif flag_type_error:
                user_input = int(input(type_error))     
                flag_type_error = False
            else:
                user_input = int(input(bounds_error))
        except:
            user_input = lower - 1      # reset to a set out of bounds value
            flag_type_error = True

    return user_input

def two_dim_array(rows: int, cols: int = None, delimiter: str = " ", item_type: int = str) -> list[list[Any]]:

    """
    Arguments: A number of rows, a number of columns, and an optional delimiter. The delimiter is set to spaces by default. To split strings into individual characters, enter a delimiter of ''. 
    You can also enter a type (str, float) in which the items will be read. The default is str.

    Raises: An AssertionError for invalid row or column counts.
    
    Returns: A 2-d list will be in the form of list[rows][cols].
    """

    output = []
    for _ in range(rows):
        row = input()           # delimiter for rows must be a \n character
        if delimiter == "":
            items = list(row)   # splits string into char array
        else:
            items = row.split(delimiter)    # otherwise split by the delimiter  
        assert cols == None or (cols != None and len(items) == cols), "Column count parameter and items in the row do not match."
        output.append(list(map(item_type, items)))

    return output


"""
This part of the module includes useful tools for printing. Currently, the following are supported:
- Clear screen
- Clear screen than print message
- 'Smooth' print a message
- Center text
- Apply a gradient to text or background
"""

import os
import re
from typing import Any, Iterable
from time import sleep
from .style import Color, Background 

# \033[ followed by a combo of numbers and ; ended with a single letter
__ANSI_PATTERN = "\033\[[0-9|;]*[a-z|A-Z]"
__CENTERED_FORBIDDEN_CHARS = "[\t]"
__UNIQUE_CHAR_LENGTHS = {
    "\b": -1
}

def clear_window() -> None:
    """
    Clears the terminal window.
    """
    if os.name == "nt":
        os.system("cls")
    else:
        os.system("clear")

def clear_print(message: Any, end: str = "\n") -> None:
    """
    Clears the terminal window and prints the prompt.
    """
    clear_window()
    print(message, end=end)

def smooth_print(message: Any, delay: int = 25, end: str = "\n") -> None:
    """
    Prints a message, smoothly.

    Arguments: A message, and delay (in milliseconds), and an optional end. 
    """

    # splits into ascii
    chars = __split_esc_chars(str(message))

    # prints each character with a delay
    for char in chars:
        if re.search(__ANSI_PATTERN, message) == None:    # delay only if ansi
            sleep(delay/1000)       
        print(char, end="")
        
    # prints the end (optional)
    print(end, end="")

def centered(message: str | Iterable[str], margin: int = 2) -> str:
    """
    Centers several lines of text.
    """
    
    assert isinstance(message, (str, Iterable)), "Messages must be iterable."
    assert isinstance(margin, int), "Margin must be an integer."

    # splits depending on the types
    if isinstance(message, str): msgs = message.split("\n")
    else: msgs = [*message]

    # preprocessing to split newlines into two and remove forbidden characters
    for i in range(len(msgs)-1, -1, -1):
        line = re.sub(__CENTERED_FORBIDDEN_CHARS, '', msgs[i])  # remove forbidden characters and strip whitespace
        if "\n" in line:
            msgs = msgs[0:i] + line.split('\n') + msgs[i+1:]
        else:
            msgs[i] = line  # different kinds of line replacement

    msgs = [__split_esc_chars(line.strip()) for line in msgs]

    # calculate the lengths of each segment
    lens = [sum([__get_effective_len(c) for c in line]) for line in msgs]
    max_len = max(lens)

    output = ""

    # adds lines and adds whitespace
    for line, ln in zip(msgs, lens):
        output += (margin + (max_len-ln)//2) * " " + "".join(line) + (margin + (max_len-ln) - (max_len-ln)//2) * " " + "\n"

    return output[:-1:] # remove last \n

def __get_effective_len(char: str) -> int:
    """
    Gets the effective length of a character for things like centering the text.
    """

    if re.match(__ANSI_PATTERN, char):
        return 0
    elif char in __UNIQUE_CHAR_LENGTHS:
        return __UNIQUE_CHAR_LENGTHS[char]
    else:
        return len(char)

def __split_esc_chars(message: str) -> list[str]:
    """
    Splits a message up into characters and escape codes.
    """

    """
    Note: "between" represents characters that are between the ANSI codes.
    """

    char_list = []

    # get codes and normal things in between
    ansi_codes = re.finditer(__ANSI_PATTERN, message)
    matches = [(match.start(), match.end(), match.group()) for match in ansi_codes]

    i = 0
    while i < len(message):

        # check if start is an ansi value
        if len(matches) > 0 and i == matches[0][0]:
            char_list.append(matches[0][2])     # add match
            i = matches[0][1]                   # move index to end of match
            matches.pop(0)                      # get rid of it
        else:
            char_list.append(message[i])        # otherwise add normal char
            i += 1
        
    return char_list

def gradient(message: str, left_rgb: tuple[int, int, int], right_rgb: tuple[int, int, int], mode: str = "background") -> str:
    """
    Returns a string with the gradient applied.
    """

    assert all([0 <= color <= 255 for color in left_rgb+right_rgb]), "Invalid RGB number entered."
    assert len(left_rgb) == len(right_rgb) == 3, "Tuple with RGB values must have a length of 3." 
    assert isinstance(message, (str, Iterable)), "Message must be a string or a list of strings representing newlines."
    assert mode in ["text", "background"], "Invalid mode."

    # tune hte rgb value depending on progress in the line: a and b are start and end values, i is the progress, and n is the max count.
    def get_value(a, b, i, n): return a + round((b-a)*(i/n))  

    # filter mode
    if mode == "text": method = Color
    elif mode == "background": method = Background
    
    # splits depending on the types
    if isinstance(message, str): messages = message.split("\n")
    else: messages = [*message]

    messages = [__split_esc_chars(line) for line in messages]
    max_len = max([len(s) for s in messages]) - 1
    rgb_vals = [tuple([get_value(left_rgb[ii], right_rgb[ii], i, max_len) for ii in range(3)]) for i in range(max_len+1)]
    output = f"{method.RESET_COLOR if method == Color else method.RESET_BACKGROUND}\n".join(            # joins the lines by a reset color/background
        ["".join([method.RGB(rgb_vals[i]) + line[i] for i in range(len(line))]) for line in messages]   # creates the line
    )

    return output


"""
All ANSI codes for colors are adapted from w3schools's ANSI colors reference:
https://www.w3schools.blog/ansi-colors-java

This is my attempt to reproduce colorama's implementation!
Combinations of styles (backgrounds, colors, and text styles) are possible.
This version includes the ability to modify RGB values of colors, set custom colors, and move the cursor.
"""

class Text:

    """
    This class stores information about editing text styles.
    """

    RESET_ALL =     "\033[0m"
    RESET_STYLE =   "\033[22m"

    BOLD =          "\033[1m"
    DIM =           "\033[2m"
    ITALIC =        "\033[3m"
    UNDERLINE =     "\033[4m"
    BLINKING =      "\033[5m"
    INVERTED =      "\033[7m"
    STRIKETHROUGH = "\033[9m"

    HIDE_CURSOR =   "\033[?25l"
    SHOW_CURSOR =   "\033[?25h"

    def MOVE_CURSOR_UP(lines: int = 1):
        """
        Returns an ANSI code to move the cursor up by <lines> lines.
        Through testing, I find that this seems to clear the lines that you go up from.
        """
        assert isinstance(lines, int) and lines >= 0, "Number of lines to move up must be a nonnegative integer."
        return f"\033[{lines}A"
    
    def MOVE_CURSOR_DOWN(lines: int = 1):
        """
        Returns an ANSI code to move the cursor down by <lines> lines.
        """
        assert isinstance(lines, int) and lines >= 0, "Number of lines to move up must be a nonnegative integer."
        return f"\033[{lines}B"
    
    def MOVE_CURSOR_RIGHT(cols: int = 1):
        """
        Returns an ANSI code to move the cursor right by <cols> columns.
        """
        assert isinstance(cols, int) and cols >= 0, "Number of columns must be a nonnegative integer."
        return f"\033[{cols}C"
    
    def MOVE_CURSOR_LEFT(cols: int = 1):
        """
        Returns an ANSI code to move the cursor left by <cols> columns.
        """
        assert isinstance(cols, int) and cols >= 0, "Number of columns must be a nonnegative integer."
        return f"\033[{cols}D"

class Background:

    """
    This class stores information about modifying the background color.
    """

    RESET_BACKGROUND =  "\033[49m"
    RESET_ALL =         "\033[0m"

    BLACK =     "\033[40m"
    RED =       "\033[41m"
    GREEN =     "\033[42m"
    YELLOW =    "\033[43m"
    BLUE =      "\033[44m"
    PURPLE =    "\033[45m"
    CYAN =      "\033[46m"
    WHITE =     "\033[47m"

    BLACK_BRIGHT =     "\033[100m"
    RED_BRIGHT =       "\033[101m"
    GREEN_BRIGHT =     "\033[102m"
    YELLOW_BRIGHT =    "\033[103m"
    BLUE_BRIGHT =      "\033[104m"
    PURPLE_BRIGHT =    "\033[105m"
    CYAN_BRIGHT =      "\033[106m"
    WHITE_BRIGHT =     "\033[107m"

    def RGB(rgb: tuple[int, int, int]) -> str:

        """
        Returns the ANSI escape code for a background color in RGB.
        """

        for var in rgb:
            assert isinstance (var, int) and 0 <= var <= 256, "Invalid RGB value!"
        return f"\033[48;2;{rgb[0]};{rgb[1]};{rgb[2]}m"
    
    def SET(name: str, rgb: tuple[int, int, int]) -> None:

        """
        Sets a custom background color given its name and a tuple containing (r, g, b) values.
        Call this color using: Background.NAME, where NAME is the entered name.
        """

        assert isinstance(name, str), "Name must be a string!"
        setattr(Background, name, Background.RGB(rgb))

class Color:

    """
    This class stores information about modifying the foreground color (color of text)
    """

    RESET_COLOR =   "\033[39m"
    RESET_ALL =     "\033[0m"

    BLACK =     "\033[30m"
    RED =       "\033[31m"
    GREEN =     "\033[32m"
    YELLOW =    "\033[33m"
    BLUE =      "\033[34m"
    PURPLE =    "\033[35m"
    CYAN =      "\033[36m"
    WHITE =     "\033[37m"

    BLACK_BRIGHT =     "\033[90m"
    RED_BRIGHT =       "\033[91m"
    GREEN_BRIGHT =     "\033[92m"
    YELLOW_BRIGHT =    "\033[93m"
    BLUE_BRIGHT =      "\033[94m"
    PURPLE_BRIGHT =    "\033[95m"
    CYAN_BRIGHT =      "\033[96m"
    WHITE_BRIGHT =     "\033[97m"

    def RGB(rgb: tuple[int, int, int]) -> str:

        """
        Returns the ANSI escape code for a text color in RGB.
        """

        for var in rgb:
            assert isinstance(var, int) and 0 <= var <= 256, "Invalid RGB value!"
        return f"\033[38;2;{rgb[0]};{rgb[1]};{rgb[2]}m"
    
    def SET(name: str, rgb: tuple[int, int, int]) -> None:

        """
        Sets a custom text color given its name and a tuple containing (r, g, b) values.
        Call this color using: Color.NAME, where NAME is the entered name.
        """

        assert isinstance(name, str), "Name must be a string!"
        setattr(Color, name, Color.RGB(rgb))

def demo():

    """
    A simple demonstration of the different colors supported.
    """

    # prints examples for each property of the color classes
    def printing_loop(var):
        for name in dir(var):
            if name[0] != "_" and sum([kw in name for kw in ["RESET", "RGB", "SET", "CURSOR"]]) == 0:
                modifier = getattr(var, name)
                print(f"{modifier}{var.__name__}.{name}{var.RESET_ALL}")

    # test cases for normal color properties
    print(f"{Text.BOLD}Colors:{Text.RESET_STYLE}")
    printing_loop(Color)
    print(f"{Text.BOLD}Backgrounds:{Text.RESET_STYLE}")
    printing_loop(Background)
    print(f"{Text.BOLD}Text Styles:{Text.RESET_STYLE}")
    printing_loop(Text)

    # test cases for rgb colors
    print(f"{Color.RGB((100, 200, 150))}Color.RGB((100, 150, 200)){Text.RESET_ALL}")
    print(f"{Background.RGB((200, 100, 100))}Background.RGB((200, 100, 100)){Text.RESET_ALL}")

    # test cases for setting values
    Color.SET("COOL_BLUE", (100, 150, 200))
    print(f"{Color.COOL_BLUE}Color.SET(\"COOL_BLUE\", (100, 150, 200)); Color.COOL_BLUE")

if __name__ == "__main__":
    demo()


"""
This part of the module holds programs that helps getting user input through menus.
"""

from typing import Iterable, Any
from .input import bounded_int

def __map_to_str(opt):

    """
    Turns the variable into a string.
    """

    if isinstance(opt, (bool, str, float, int)):
        return str(opt)
    try:
        return opt.__name__
    except:
        raise AssertionError("Option type unable to be converted to string using __name__.")

def numbered_menu(options: Iterable[Any], beginning_prompt: str = None, 
                  selection_prompt: str = None, spacing: int = 4, return_number: bool = False) -> Any:

    """
    Creates a numbered menu, in the following format:
    ```
    '''
    <prompt>
        1. Option
        2. Option
        3. Option
    Please enter a number between {lower} and {upper}:
    '''
    ```
    Arguments: The list of options to enter, which can include things like classes, functions, or simple datatypes. An optional prompt, and an optional spacing value (how many spaces the options are intended by). Additionally, you can enter a return_number parameter in order to return the number rather than its associated value.

    Raises: An AssertionError for incompatible types.

    Returns: The choice selected by the user.
    """

    assert isinstance(options, Iterable), "Options must be iterable!"
    assert len(options) >= 2, "There should be at least two values to select from!"
    assert isinstance(spacing, int) and spacing >= 0, "Spacing must be a non-negative integer!"

    str_options = list(map(__map_to_str, options))

    if beginning_prompt:
        print(beginning_prompt)

    for index, opt in enumerate(str_options):
        print(f"{' '*spacing}{index+1}. {opt}")
    
    choice = bounded_int(1, len(str_options), prompt=selection_prompt)

    if return_number:
        return choice
    
    return options[choice - 1]

def list_menu(options: Iterable[Any], beginning_prompt: str = None, 
              selection_prompt: str = "Enter your choice: ",
              error_prompt: str = "Invalid choice, enter a valid one: ",
              spacing: int = 4, selector: str = '-') -> Any:

    """
    Creates a numbered menu, in the following format:
    ```
    '''
    <prompt>
        - Option
        - Option
        - Option
    Please enter yor option:
    '''
    ```
    Arguments: The list of options to enter, which can include things like classes, functions, or simple datatypes. An optional prompt, and an optional spacing value (how many spaces the options are intended by).

    Raises: An AssertionError for incompatible types.

    Returns: The choice selected by the user.
    """
    
    assert isinstance(options, Iterable), "Options must be iterable!"
    assert len(options) >= 2, "There should be at least two values to select from!"
    assert isinstance(spacing, int) and spacing >= 0, "Spacing must be a non-negative integer!"
    assert isinstance(selector, str), "List element introducer must be a string!"

    str_options = list(map(__map_to_str, options))

    if beginning_prompt:
        print(beginning_prompt)

    for opt in str_options:
        print(f"{' '*spacing}{selector} {opt}")

    choice = input(selection_prompt)

    while choice.strip() not in str_options:
        choice = input(error_prompt)

    return options[str_options.index(choice)]


from setuptools import setup, find_packages
import pynterface

with open("README.md", "r") as f:
    description = f.read()

setup(
    name="pynterface",
    version=pynterface.__version__,
    author=pynterface.__author__,
    author_email='singhvi.vivaan@gmail.com',
    description='Terminal-Based Printing Tools!',
    long_description=description,
    long_description_content_type="text/markdown",
    license='MIT',
    packages=find_packages(),
    python_requires='>=3.8',
    include_package_data=True,
    project_urls={
        "Documentation": "https://pynterface.readthedocs.io/en/latest/"
    }
)